import serial
import numpy as np
from ctypes import *
from enum import Enum
from dataclasses import dataclass

BYTES_PER_BIT_TRIPLET = 8
TRIPLET_ARRAY = [0x5b, 0x1b, 0x53, 0x13, 0x5a, 0x1a, 0x52, 0x12]

RED_MAX_CURRENT = .015
GREEN_MAX_CURRENT = .015
BLUE_MAX_CURRENT = .015

MAX_COLOR_CODE = 256.0

class SetColors(Enum):
    Red = 0, 
    Green = 1,
    Blue = 2,
    Yellow = 3,
    Cyan = 4,
    Magenta = 5,
    White = 6, 
    Black = 7

@dataclass
class Color:
    red: np.uint8
    blue: np.uint8
    green: np.uint8

# 3-bit color lookup table for WS2812B
bitTriplets = np.array(TRIPLET_ARRAY, dtype=np.uint8)

# Set predetermined color
def createSetColor(color: SetColors=SetColors.Black, intensity: int=0) -> Color:
        if color is not None:
            if color == SetColors.Red:
                red = np.uint8(intensity)
                blue = 0x00
                green = 0x00
            elif color == SetColors.Green:
                red = 0x00
                blue = 0x00
                green = np.uint8(intensity)
            elif color == SetColors.Blue:
                red = 0x00
                blue = np.uint8(intensity)
                green = 0x00
            elif color == SetColors.Yellow:
                red = intensity
                blue = 0x00
                green = intensity
            elif color == SetColors.Cyan:
                red = 0x00
                blue = np.uint8(intensity)
                green = np.uint8(intensity)
            elif color == SetColors.Magenta:
                red = np.uint8(intensity)
                blue = np.uint8(intensity)
                green = 0x00
            elif color == SetColors.White:
                red = np.uint8(intensity)
                blue = np.uint8(intensity)
                green = np.uint8(intensity)
            elif color == SetColors.Black:
                red = 0x00
                blue = 0x00
                green = 0x00
            else:
                raise ValueError('Invalid color')
            
        ret = Color(red, blue, green)
        return ret

def packColor(color: Color) -> np.uint32:
    val = np.uint32(0)
    val = (color.green << 16) | (color.red << 8) | color.blue
    return val

def generateByteData(color: list[Color], colorData: np.array) -> None:
    
    # Check lengths
    if len(color) * BYTES_PER_BIT_TRIPLET > len(colorData):
        print("Error: Not enough space to pack colors into colorData array")
        return

    # Pack color(s)
    offset = 0
    for c in color:
        # Pack the color 
        packed = packColor(c)

        # Generate byte data for color
        for i in range(BYTES_PER_BIT_TRIPLET):
            colorData[i + (offset * BYTES_PER_BIT_TRIPLET)] = bitTriplets[(packed >> ((BYTES_PER_BIT_TRIPLET - i - 1) * 3)) & 0x07]

        # Increment offset into colorData
        offset += 1
    return

def calculateCurrentDraw(colorList: list[Color]) -> tuple[float]:
    redCurrent = 0.0
    greenCurrent = 0.0
    blueCurrent = 0.0

    for c in colorList:
        redCurrent += (float(c.red)/MAX_COLOR_CODE) * RED_MAX_CURRENT
        greenCurrent += (float(c.green)/MAX_COLOR_CODE) * GREEN_MAX_CURRENT
        blueCurrent += (float(c.blue)/MAX_COLOR_CODE) * BLUE_MAX_CURRENT

    total = redCurrent + greenCurrent + blueCurrent
    return total, redCurrent, greenCurrent, blueCurrent

class WS2812():
    def __init__(self, comPort, ledLen, maxCurrent=.500):

        # Store length of LEDs and max current
        self.__serDev = None
        self.__ledLen = ledLen
        self.__maxCurrent = maxCurrent
        
        # Open serial port and connect with custom settings 
        # to drive WS2812 strip 
        try:
            self.__serDev = serial.Serial(comPort, 3000000, serial.SEVENBITS, serial.PARITY_NONE, serial.STOPBITS_ONE, timeout=.200)
        except Exception as e:
            print(e)
            self.__serDev = None
            return
        
        # Create memory buffers for colors and generated data 
        self.__colorBuff = [None] * ledLen
        self.__colorDataBuff = np.zeros(ledLen * BYTES_PER_BIT_TRIPLET, dtype=np.uint8)

        # Set them all to black:
        for i in range(len(self.__colorBuff)):
            self.__colorBuff[i] = createSetColor()

        # Update pattern and send it out 
        self.updateStrip()

    def __getitem__(self, i:int):
        if self.__ledLen <= i:
            raise IndexError('Index out of range')
        return self.__colorBuff[i]
    
    def __setitem__(self, i:int, item: Color):
        if self.__ledLen <= i:
            raise IndexError('Index out of range')        
        self.__colorBuff[i] = item

    def setAll(self, color: Color) -> None:
        for i in range(self.__ledLen):
            self.__colorBuff[i] = color

    def updateStrip(self):
        
        # Measure maximum current draw of strip
        totalI, redI, greenI, blueI = calculateCurrentDraw(self.__colorBuff)
        #print(f"Total Current: {totalI:.3f} A, Red: {redI:.3f} A, Green: {greenI:.3f} A, Blue: {blueI:.3f} A")

        # Update colorDataBuff
        generateByteData(self.__colorBuff, self.__colorDataBuff)

        # Send it out 
        if self.__serDev is not None:
            self.__serDev.write(self.__colorDataBuff)        

    def close(self):
        """
        Close serial port 
        """
        if self.__serDev is not None:
            self.__serDev.close()